#!/bin/bash
# Copyright (c) Daniel Sheffield 2024 - 2025

set -euo pipefail
if [ "$#" == "0" ]
then
	echo recall
	exit 0
fi

. "${BASH_SOURCE[0]%/*}"/helpers/common.sh
. "${BASH_SOURCE[0]%/*}"/helpers/awk.sh
. "${BASH_SOURCE[0]%/*}"/helpers/find.sh
. "${BASH_SOURCE[0]%/*}"/helpers/fzf.sh

stderr_deps=/dev/null
# NOTE: uncomment to see err output from `which` (if any)
#stderr_deps='/proc/self/fd/2'
check_deps 3>&2 2>"$stderr_deps" 1>/dev/null

: "${RECALL_FZF_ENABLE:=true}"
: "${RECALL_ISELECT_ENABLE:=true}"
: "${RECALL_VIPE_ENABLE:=true}"
INCLUDE_STDERR=false
EXIT_SUCCESS=false
LIST_MODE=false
CMD_STRING=false
ORS='\n'

usage () {
	cat <<EOF
Recall prints the last captured output from the specified PROG and exits with
the captured exit status.

  usage: "${0##*/} [ -l ] [ -n NUM ] PROG [ ARGS ... ]

  examples:
    - Print the output of last captured find command and exit with captured
      exit status

      ${0##*/} find

    - Print the output of last captured find command and exit with success

      ${0##*/} -z find

    - List paths to all captured commands

      ${0##*/} -l

    - List paths to all captured find commands

      ${0##*/} -l find

    - List at most 10 paths to all captured find commands

      ${0##*/} -l -n 10 find

    - List paths null delimited for pipe (in case of paths with new line char)

      ${0##*/} -l -0 find | xargs -r -0 ls -lahd

    - List last 30 command strings (bash escaped)

      ${0##*/} -lsn 30
EOF
}

OPTSTRING=":ehln:sz0"
parse_opt () {
	case "${opt}" in
		e)
			INCLUDE_STDERR=true
		;;
		h)
			usage
			exit 0
		;;
		l)
			LIST_MODE=true
		;;
		n)
			if [ "${OPTARG:-}" ] && [[ "$OPTARG" =~ [0-9]+ ]] && (( OPTARG > 0 ))
			then
				NUM="$OPTARG"
			else
				echo "-n paremeter must be an integer greater than zero" >&2
				echo >&2
				usage >&2
				exit 1
			fi
		;;
		z)
			EXIT_SUCCESS=true
		;;
		s)
			CMD_STRING=true
		;;
		0)
			ORS='\0'
		;;
		\?)
			echo "Unknown switch: -$OPTARG" >&2
			echo >&2
			usage >&2
			exit 1
		;;
	esac
}

validate () {
	# exit if vars are not valid
	# ie, set any dynamic defaults here
	if ! "$LIST_MODE" && [ "${NUM:=1}" ] && ! (( NUM == 1 ))
	then
		echo "-n != 1 can only be specified with -l" >&2
		echo >&2
		usage >&2
		exit 1
	fi
	if ! "$LIST_MODE" && [ "$#" == "0" ]
	then
		echo "PROG must be provided" >&2
		echo >&2
		usage >&2
		exit 1
	fi
}

while getopts "${OPTSTRING}" opt
do
	parse_opt
done
shift $((OPTIND-1))

validate "$@"

sort_rw0 () {
	# NOTE: sort by program, or timestamp
	if [ "$1" = "prog" ]
	then
		sort -zrV
	elif [ "$1" = "ts" ]
	then
		awk \
			-v FS='/' -v RS='\0' \
			-v OFS='/' -v ORS='\0' \
			'{print $NF,$0}' \
		| sort -zrV \
		| awk_rw0_strip_first_path_part \
		:
	else
		echo "unsupported sort option: $1" >&2
		false
	fi
}

preview_rw0 (){
	header_size="$1"
	if "$RECALL_FZF_ENABLE" && which fzf 1>/dev/null 2>&1
	then
		fzf_rw0_inplace_preview "$header_size"
	elif "$RECALL_ISELECT_ENABLE" && which iselect 1>/dev/null 2>&1
	then
		while read -rd "" line
		do
			echo -n "#"
			head -n1 "$line"/info
			echo "<s>$line"
		done \
		| iselect -fmt recall -n history \
		| awk \
			-v RS='\n' \
			-v ORS='\0' \
			'{print $0}' \
		&& :
	elif "$RECALL_VIPE_ENABLE" && which vipe 1>/dev/null 2>&1
	then
		while read -rd "" line
		do
			echo -n "#"
			head -n1 "$line"/info
			echo "$line"
		done \
		| vipe \
		| awk \
			-v RS='\n' \
			-v ORS='\0' \
			'/^[^#]/ {print $0}' \
		&& :
	else
		cat
	fi
}

list_r0 (){
	ors="$1"
	if ! "$CMD_STRING"
	then
		awk_rw0_strip_first_path_part \
		| awk_r0_add_prefix "$ors" "$LOG_ROOT" \
		&& :
	else
		awk_r0_add_suffix '\0' 'info' \
		| xargs -r0n1 head -n1 \
		| awk \
			-v RS='\n' \
			-v ORS="$ors" \
			'{print $0}' \
		&& :
	fi
}

if "$LIST_MODE"
then
	# TODO: make sort option configurable
	sort_opt="ts"
	find_name=()
	if [ "${1:-}" ]
	then
		slugs=( "$1" )
		if [ "${2:-}" ]
		then
			slugs+=( "$2" )
		fi
		query="$(bash -ac 'IFS=/; echo "$*"' "recall-bash" "${slugs[@]}")"
		find_name=( -wholename "**/$query/**" )
	fi
	(
		cd "$LOG_ROOT"
		find_w0_leaf_directories "${find_name[@]}"\
		| sort_rw0 "$sort_opt" \
		| head -zn "${NUM:--0}" \
		| preview_rw0 "2" \
		| list_r0 "$ORS" \
		&& :
	)
else
	PROG="$1"
	SUBPROG="$(get_subprog "$@")"

	LOG_DIR="$LOG_ROOT"/"$PROG"/"$SUBPROG"
	LOG_DIR="$LOG_DIR"/"$(cd "$LOG_DIR" && find . -mindepth 1 -maxdepth 1 -print0 | sort -zV | tail -zn1 | xargs -r -0 echo)"
	OUT="$LOG_DIR"/stdout
	ERR="$LOG_DIR"/stderr
	# shellcheck disable=SC2034
	DAT="$LOG_DIR"/dat
	INF="$LOG_DIR"/info
	# shellcheck disable=SC1090
	. <(tail -n +2 "$INF") # first line is commandline
	"$EXIT_SUCCESS" && status=0
	"$INCLUDE_STDERR" && cat "$ERR" >&2
	cat "$OUT"
	exit "$status"
fi
